/* LUGS Luxus Gasumschalter fuer AEB-Autogas-Steuergeraete          */
/*                                                                  */
/* (c) C.Schirp  released for non-commercial use                    */
/* LCD-Routinen (c) Holger Klabunde mit Anpassungen von mir         */
/*                                                                  */
/* V1.0 Erstausgabe (ohne Taster, ohne Pieper, ohne Powerdown)      */
/* V1.1 Taster, Hintergrundbeleuchtung, Pieper                      */
/* V1.2 Bug in do_taste() bereinigt                                 */
/* V1.3 rx_code und disp_code getrennt                              */
/* V1.4 Taste wird nicht mehr per SW ausgewertet, RA5=Input         */
/* V1.5 Startzeit LOW fuer TX_Pin hinzu                             */
/* V1.6 Warteroutine auf 1ms getrimmt, Tanksymbole, LCD schneller   */
/*       LCD Power vom uC geschaltet                                */
/*                                                                  */

#include <c:\projekte\picsel\cc5\16F628.h>
#include <c:\projekte\picsel\cc5\int16CXX.h>
#include <c:\projekte\picsel\cc5\inline.h>

// interner Reset-Generator
#pragma config = 0x3F18
// int Res und Brown Out
//#pragma config = 0x3F58

#define LCD_RS    0b.00010000             //RS-Pin ist Bit4, LCD Register Select
#define LCD_CS    0b.01000000             //E-Pin ist Bit6, LCD Chip Select

#define BUZZER   	RB2
#define BACKLIGHT   	RB5
#define TX_PIN 		RB1
#define LCD_POWER       RB3

#pragma char PORTQ @ PORTA               //PORT auswhlen, DB7-Bit7, DB6-Bit6, DB5-Bit5, DB4-Bit4

#define LCDCS_ON() { NOP;  NOP; lcdport|=LCD_CS; PORTQ=lcdport; }
#define LCDCS_OFF() { NOP;  NOP; lcdport&=~LCD_CS; PORTQ=lcdport; }
#define LCDRS_ON() { NOP;  NOP; lcdport|=LCD_RS; PORTQ=lcdport; }
#define LCDRS_OFF() { NOP; NOP; lcdport&=~LCD_RS; PORTQ=lcdport; }

#define Wait() Delay1ms(3)

//Kommandos fr das LCD-Display
#define CLEAR_DISPLAY 0x01
#define OFF_DISPLAY   0x08
#define CURSOR_OFF    0x0C
#define CURSOR_ON     0x0E
#define CURSOR_BLINK  0x0F
#define SHIFT_LEFT    0x18
#define SHIFT_RIGHT   0x1C
#define CGRAM_LEN	56  //Anzahl zu bertragender Zeilen ins CGRAM

#define K_BLINK	50 	// Zeitkonstante fr blinkende Texte
#define K_BUZZ 5	// Zeitkonstante fr Buzzer On


/* Variablen */

bank0 uns8		lcdport;
bank0 uns8 		Blink:1;	// Bitcounter fr LCD-Blinken
bank0 uns8		Buzz:1;		// Bitcounter fr Buzzer
bank0 uns8      	Blink_rq:1;	// Statustext Blinkend ausgeben
bank0 uns8		Buzz_rq:1;	// Beep bei Empfang
bank0 uns8		Buzz_rq_alt:1;
bank0 uns8		Buzz_rpt:1;	// Piepen wiederholen
bank0 uns16  		rx_code;	// Empfangener Code 0000=ungueltig
bank0 uns16		disp_code;	// Angezeigter, dekodierter Code
bank0 uns8		fuellstand;	// Fuellstand
bank0 uns8		status;		// Status
bank0 uns8		rx_merkbit:1;	// 1 if RX is Hi BITVARIABLE!
bank0 uns8		Text[9];	// Speicher fr erste LCD-Zeile incl Nullterminierung
bank0 uns8		Text2[9];	// Speicher fr zweite LCD-Zeile incl Nullterminierung
bank0 uns8		tmr0cpy;
bit RX_PIN     	@ PORTB.0;	// INT_IN

const unsigned char chardefs[] = {
//Adresse 1 im Character RAM
// Doppleklotz fuer Tankanzeige
0x00, 0x00, 0x00, 0x00, 0x00, 0x1b, 0x1b, 0,
0x00, 0x00, 0x00, 0x00, 0x1b, 0x1b, 0x1b, 0,
0x00, 0x00, 0x00, 0x1b, 0x1b, 0x1b, 0x1b, 0,
0x00, 0x00, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0,
0x00, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0,
// Leer-Symbol
0x11, 0x11, 0x11, 0x11, 0x11, 0x1F, 0x1F, 0x0,
// Voll-Symbol
0x11, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x0,

};

void ClrText(void);
/*------------------------------------------------------------------*/
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*------------------------------------------------------------------*/


#pragma origin 4
interrupt int_server( void)
{
    int_save_registers    // W, STATUS (and PCLATH)

    if (INTF)
    {
        /* RB0/INT interrupt */
        rx_merkbit = !RX_PIN;
        INTF = 0;
        if (rx_merkbit)     // request interrupt on next edge (up or down)
           INTEDG = 1; // if 1, then interrupt on rising edge
        else
           INTEDG = 0; // is part of OPTION_REG, is 1 on reset

        /* finish and restart TMR0 */
        tmr0cpy = TMR0;
        TMR0 = 0;
        T0IF = 1; // end polling wait-loop, similar to TRM0 wraparound
    }

   /*
     NOTE: GIE is AUTOMATICALLY cleared on interrupt entry and set
           to 1 on exit (by RETFIE). Setting GIE to 1 inside the
           interrupt service routine will cause nested interrupts
           if an interrupt is pending. Too deep nesting may crash
           the program !
   */

    int_restore_registers // W, STATUS (and PCLATH)
}


/*------------------------------------------------------------------*/
/*   II  N   N  II TTTTT                                            */
/*   II  NN  N  II   T                                              */
/*   II  N N N  II   T                                              */
/*   II  N  NN  II   T                                              */
/*   II  N   N  II   T                                              */
/*------------------------------------------------------------------*/


void init ()
{
	/* PORT A, B Config */
	CMCON = 0x07;		// kein Komparator
	
	TRISB = 0b0000.0001; 	// Ausgaenge: 1,2,5 Eingaenge: 0
	TRISA = 0b0010.0000;	// LCD-Port: PA output, ra5 input
	
	PORTB = 0b0010.0000;	// HINT: erst Direction, dann Wert!
	PORTA = 0x00;
	/*sonstiges*/
	T1CON = 0x00;		// Timer disabled
	OPTION_REG = 0x04;/* pull-up portB, falling edge int, no WDT prescaler, prescale=32 */


	/* Variablen initialisieren */
	rx_merkbit = 	0;
	Blink_rq=	0;
	Buzz_rq=	0;
	Buzz_rq_alt = 	0;
	

	INTCON = 0; 			// reset and disable all interrupts by default
	INTE = 	1;
	GIE = 	 1;				// Enable Interrupts
	
	disp_code = 0;
	rx_code = 0;
	Text[8]=0;
	Text2[8]=0;
	
	ClrText();
}

/*------------------------------------------------------------------*/
/*   LL    CCCC DDD                                                 */
/*   LL   CC    DD DD                                               */
/*   LL   CC    DD  DD                                              */
/*   LL   CC    DD DD                                               */
/*   LLLLL CCCC DDD                                                 */
/*------------------------------------------------------------------*/

void Delay1ms(uns16 ms)
{
#pragma updateBank 0 /* OFF, all cores */

	uns8 count;
 do
 {
 	ms--;	
	for (count = 0; count < 0x87; count++) // F0
	{
		NOP;	
	}
}
while (ms);

#pragma updateBank 1 /* ON, all cores */
}

void Delay100u(uns16 us)
{
#pragma updateBank 0 /* OFF, all cores */

	uns8 count;
 do
 {
 	us--;	
	for (count = 0; count < 0x10; count++) // F0
	{
		NOP;	
	}
}
while (us);

#pragma updateBank 1 /* ON, all cores */
}

//*******************************************
// Schreibt ein Datenbyte in die LCD-Anzeige
// Schreibt das Byte in zwei 4 Bit Nibbles
//*******************************************

void LCDWriteByte (unsigned char data)
{
 unsigned char buff;
 lcdport= PORTQ; 		//alten Wert merken
 LCDCS_OFF();
 buff=data>>4;   		//High Nibble des Bytes
 buff&=0x0F;			//Rest lschen
 lcdport=lcdport&0xF0;		//Hier D0-D3 lschen zum
 lcdport|=buff;			//reinodern vom High Nibble
 lcdport|=LCD_CS;
 LCDCS_ON();	      		//High Nibble schreiben

 LCDCS_OFF();
 buff=data&0x0F;        	//Low Nibble des Bytes
 lcdport=lcdport&0xF0;		//D0-D3  lschen zum
 lcdport|=buff;			//reinodern vom Low Nibble
 LCDCS_ON();	      		//Low Nibble schreiben
 LCDCS_OFF();
 NOP;
 NOP;
}

//*****************************************
// Commando ans Display senden
//*****************************************
void LCDCommand(unsigned char mode)
{
 LCDRS_OFF(); 			//Controlregister
 LCDWriteByte(mode);
 LCDRS_ON(); 			//Zurck auf Datenregister
}

//*****************************************
// Lscht die LCD Anzeige
// Cursor steht dann auf Position 1,Zeile 1
// Gilt auch fr mehrzeilige Anzeigen
//*****************************************
void LCDCls(void)
{
 LCDRS_OFF(); //Controlregister
 LCDWriteByte(0x01); //CLS
 LCDRS_ON(); //Umschalten auf Datenregister
 Wait();
}
//**************************************************
//Schreibt einen Text in die LCD-Anzeige
//**************************************************
void LCDString(const char *__putstr)
{
#pragma updateBank 0 /* OFF, all cores */
 char __ps;

 __ps = *__putstr;

  while(__ps>0)
   {
    __putstr++;
    if (__ps== 0) break;
   LCDWriteByte(__ps);
    __ps = *__putstr;
   }
#pragma updateBank 1 /* ON, all cores */
}
//****************************************************
// Springt auf Zeile x und Position y der LCD-Anzeige
// Gltig fuer 2 Zeilen : 0,1.
//****************************************************
void LCDPos(unsigned char zeile,unsigned char position)
{
#pragma updateBank 0 /* OFF, all cores */
	 unsigned char buff;
	
	// 0x80 wegen Display Adress Set.
	// Erste Position ist bei mir  0.
	
	if (zeile)
	{
		buff=0xC0+position;
	}
	else
	{
		buff=0x80+position;
	}
	
	 LCDCommand(buff);
#pragma updateBank 1 /* ON, all cores */
}


//*********************************************
//Stellt die Funktionsweise der LCD Anzeige ein
//*********************************************

void LCDInit(void)
{
 int i;
//Software Reset laut Hitachi Datenblatt beim KS0066 nicht implementiert
 Delay1ms(30); //Mglichen Power On Reset abwarten 30ms reichen ab 4,5V
 
 LCDRS_OFF();  //Control Register
 LCDCS_OFF();  //LCD-Anzeige abwhlen


 /*
 for(i=0; i<3; i++)
  {
   lcdport=lcdport&0xF0;//D0-D3 lschen
   lcdport|=0x03;	//Wert reinodern
   PORTQ=lcdport;	//Auf den Port legen
   LCDCS_OFF(); //LCD-Anzeige abwhlen
   NOP;
   LCDCS_ON();	//High Nibble schreiben
   Delay1ms(5); //Ein bichen warten auf die Anzeige
  } 
  */
//Ende Software Reset


//Der erste Schreibzugriff nach dem Reset ist ein 8 Bit
//Zugriff!! Also nur einmal schreiben, nicht Low-High Nibble.

//Um die 4Bit Operation einzuschalten
 lcdport=0x02;	//4 Bit Operation	

 PORTQ=lcdport;	//Auf den Port legen
 NOP;

 LCDCS_ON();	//High Nibble schreiben
 NOP;
 NOP; 
 LCDCS_OFF();
 Delay100u(1);
 LCDWriteByte(0x2C);    //4 Bit Operation,2 Line,5x7 Font  5
 Delay100u(1);
 LCDWriteByte(0x14); 	//Cursor Move,Right Shift
 Delay100u(1);

 LCDWriteByte(0x0C); 	//Display on, Cursor off ,Cursor no blink
 Delay100u(1);

 LCDWriteByte(0x06); 	//Increment, Display Freeze
 Delay100u(1);
 LCDWriteByte(0x02); 	//Cursor Home
 Delay100u(1);
 LCDRS_ON();		//LCD-Datenregister

 LCDCommand(CLEAR_DISPLAY);
 Wait();
 
 //Die eigenen Character ins CGRAM schreiben

//Der interne Adresszhler des LCD-Displays luft
//nach jedem Schreibzugriff automatisch hoch

 LCDCommand(0x48);//CG_RAM Adresse(1) setzen

 for(i=0; i<CGRAM_LEN; i++) //8 Codes pro Zeichen bertragen
  {
   LCDWriteByte(chardefs[i]);
  }

 LCDCommand(0x80); //DD_RAM Adresse 0 setzen
}


/*------------------------------------------------------------------*/
/*  Konvertiert 16 Bit Binaer zu Hexadezimal-ASCII 4-stellig        */
/*------------------------------------------------------------------*/
void Dez4Hex(uns16 zahl)
{

#pragma updateBank 0 /* OFF, all cores */
	char  count, help;

	for (count = 0; count < 4; count++)
	{
		
		help = zahl & 0x000F;
		
		if (help > 9)
		{
			Text[3 - count] = help + 0x37;
		}
		else
		{
			Text[3 - count] = help + 0x30;
		}
		zahl >>=4;
	}
#pragma updateBank 1 /* ON, all cores */
}


void ClrText(void)
{

#pragma updateBank 0 /* OFF, all cores */
#asm
	movlw 32
	movwf Text
			;	Text[1] = ' ';
	movwf Text+1
			;	Text[2] = ' ';
	movwf Text+2
			;	Text[3] = ' ';
	movwf Text+3
			;	Text[4] = ' ';
	movwf Text+4
			;	Text[5] = ' ';
	movwf Text+5
			;	Text[6] = ' ';
	movwf Text+6
			;	Text[7] = ' ';
	movwf Text+7

#endasm
#pragma updateBank 1 /* ON, all cores */
}

void ClrText2(void)
{

#pragma updateBank 0 /* OFF, all cores */
#asm
			
	movlw 32
			;	Text2[0] = ' ';
	movwf Text2
			;	Text2[1] = ' ';
	movwf Text2+1
			;	Text2[2] = ' ';
	movwf Text2+2
			;	Text2[3] = ' ';
	movwf Text2+3
			;	Text2[4] = ' ';
	movwf Text2+4
			;	Text2[5] = ' ';
	movwf Text2+5
			;	Text2[6] = ' ';
	movwf Text2+6
			;	Text2[7] = ' ';
	movwf Text2+7

#endasm
#pragma updateBank 1 /* ON, all cores */
}

/*------------------------------------------------------------------*/
/*   DDD    EEEEE  CCC                                              */
/*   DD DD  E     C                                                 */
/*   DD DD  EEE   C                                                 */
/*   DD DD  E     C                                                 */
/*   DDD    EEEEE  CCC                                              */
/*------------------------------------------------------------------*/

// 4MHz crystal, 1MHz cycle time, TMR0 prescale=32: timer units of 32 usec.
// TRM0 expires at 8 msec.

uns8 tmr0wait( void)
{
	tmr0cpy = 0xff; // potential longest waiting time, reduced by rx_merkbit
	while (!T0IF) // wait for TMR0 wrap-around or rx_merkbit
		;

	T0IF = 0;
	return tmr0cpy;
}

void bogus_code( void)
{
#pragma updateBank 0 /* OFF, all cores */

	// just eat rx_merkbit input, as long as we have fast pulses
	
	uns8 t;

	t = 0;
	while (t < 200)
	   t = tmr0wait();

	rx_code = 0x0000L; // bogus code;
	
#pragma updateBank 1 /* ON, all cores */
}

void read_code(void) 
{
#pragma updateBank 0 /* OFF, all cores */
	// 1 is sequence dark,light:  2x 900us == 2x28
	// 0 is sequence light,dark:  2x 900us == 2x28
	// frame is: one,one,toggle,11xdata
	unsigned char t;
	unsigned char i;

	t = 0;
	i = 0;
	rx_code = 1L;
	while (t < 0xa0)
	{
	   
	   t=tmr0wait(); // low_phase
	   t=tmr0wait(); // high_phase
	   
	   if( t  > 0x77 && t  < 0x84)  // High-Bit 4ms
	   {
	   	rx_code <<=1;
	   	rx_code.0 = 1;
	   	i++;
	   }
	   else if( t  > 0x37 && t  < 0x44) // Low-Bit 2ms
	   {
	   	rx_code <<=1;
	   	rx_code.0 = 0;
	   	i++;
	   }

	}

	// code in Ordnung, dann bernehmen
	if (i != 13)
	{
		rx_code = 0x0L; // bogus
	}
	else
	{
		disp_code = rx_code;
	}


#pragma updateBank 1 /* ON, all cores */	
}

void get_code( void)
{

#pragma updateBank 0 /* OFF, all cores */

	uns16 ton;
	uns16 toff;

    	toff = (uns16)tmr0wait(); // measure first low
	ton = (uns16)tmr0wait(); // following first high


    if ((ton  > 0x77) && (ton  < 0x84) &&  // nominal 81 Start Bit immer HIGH!
        (toff > 0x58) && (toff < 0x64))  // nominal 61
	{
		read_code();
	}
	else
	{
		bogus_code();
	}

#pragma updateBank 1 /* ON, all cores */	
}

void Disp_Status(void)
{
#pragma updateBank 0 /* OFF, all cores */

	LCDPos(0,0); 
	Buzz_rq=0;
	Buzz_rpt=0;
	Blink_rq = 0;
	BACKLIGHT = 1;
	switch (status)
	{
		case 0:		if (disp_code.6)
				{
					LCDString("L E E R "); Blink_rq = 1; Buzz_rq=1; Buzz_rpt=1;
				}
				else
				{
					LCDString("AUS     "); BACKLIGHT = 0; 
				}
				break;
		case 1: 	LCDString("-WARTEN-"); Blink_rq = 1; break;
		case 2: 	LCDString("GAS     "); Buzz_rq=1; break;
		case 3: 	LCDString("BENZIN  "); break;
		default: 	ClrText(); // Code ausgeben
				Dez4Hex(disp_code);
				LCDString(Text);
				break;			
	}
	
	
#pragma updateBank 1 /* ON, all cores */
}

void decode(void)
{
#pragma updateBank 0 /* OFF, all cores */

	uns8 count;
	
	/* Status */

	status = disp_code.low8;
	status >>=3;
	status &= 0x07;
	
	Disp_Status();
	
	/* Fuellstand berechnen und ausgeben */
	ClrText2();
	
	// Fuellstand nur ausgeben wenn Status != 0
	if (status)
	{
		Text2[0] = 6;
		Text2[6] = 7;
	
		fuellstand = disp_code.low8;
		fuellstand &= 0x07;
		
		// Auf max. begrenzen
		if (fuellstand > 4)
			fuellstand = 4;
		
		fuellstand++;
		
		// Balken erzeugen
		for (count = 1; count <= fuellstand; count++)
		{
			Text2[count]=count;
		}
	}
	
	LCDPos(1,0);         
	LCDString(Text2);
	
	
	
#pragma updateBank 1 /* ON, all cores */	
}



/*------------------------------------------------------------------*/
/*     M     M    A    II  N   N                                    */
/*     MM   MM   A A   II  NN  N                                    */
/*     M M M M  A   A  II  N N N                                    */
/*     M  M  M  AAAAA  II  N  NN                                    */
/*     M     M  A   A  II  N   N                                    */
/*------------------------------------------------------------------*/

void main (void)
{
#pragma updateBank 0 /* OFF, all cores */

	uns8 tmp;
	uns8 blink_cnt = K_BLINK;
	uns8 buzz_cnt = 0;

	init();
	
	// ca. 18ms fuer Stabilisierung der 5 V, dann TX High
	// im Original-16C505 erledigt das der DRT = Device Reset Timer
	Delay1ms(10);
	LCD_POWER = 1;
	
	Delay1ms(8);
	TX_PIN = 1;
	
	LCDInit();	  // LCD Initialisieren
		
	// LCDPos(0,0); nicht notwendig, steht nach CLS schon auf 0,0  

	while (1)
	{
		tmp = tmr0wait(); // wrap-around or interrupt, wrap-around takes 8ms


	    if (!RX_PIN) //came out of wait by interrupt: RX = Lo => start of new code
		{
			
			get_code();
			
			if (disp_code != 0)
			{
				decode();
				// Wenn Buzzer-Repeat, Bedingungen zuruecksetzen,
				// bei der nchsten Botschaft gibts wieder nen BEEP
				if ( Buzz_rpt )
				{
					Buzz_rq_alt = 0;
					Buzz_rq = 0;
				}
			}
			else
				Blink_rq = 0; // Kein Code empfangen -> nicht blinken!

		}
		
		// Blinkende Texte darstellen
		
		blink_cnt--;
		
		// erste Zeile blinken lassen, wenn Blink_rq
		if (!blink_cnt && Blink_rq)
		{
			Blink ^=1;	// Blinker weiterzaehlen
			
			LCDPos(0,0);
			
			if (Blink)
			{	
				LCDString("        ");
			}
			else
			{
				Disp_Status();
			}
			
			blink_cnt = K_BLINK;
		}
		else if (!Blink_rq) // ansonsten Zeitkonstante setzen
		{
			blink_cnt = K_BLINK;
		}
		
		// Buzzer
		if (Buzz_rq && !Buzz_rq_alt)
		{
			if (buzz_cnt < 4)
			{
				BACKLIGHT = 0;
				BUZZER = 1;
				buzz_cnt++;
			}	
			else
			{
				Buzz_rq_alt =1;
				BUZZER = 0;
				BACKLIGHT = 1;
			}
		}
		else
		{
			Buzz_rq_alt = Buzz_rq;
			buzz_cnt=0;
		}
		
	}
#pragma updateBank 0 /* OFF, all cores */
}
		